/*
 * Scan sql statement text and substitute fixed texts for constats of
 * variouse types:
 *
 *	numeric:	<number>
 *	string:		<string>
 *	datetime:	<datetime>
 *	interval:	<interval>
 *
 * All text is converted into lowercase. Whitespace is reduced to one blank
 * and a \n is outputtet if the line would become too long. All sql comments
 * (-- and { ... }) are removed.
 *
 * The purpose of all this is to output the same text for two statements with
 * the same structure but different constants, whitespace etc.
 *
 * Usage: scansql [input-file [output-file] ] 
 *
 *        input-file and output-file can be '-' for stdin and stdout
 */

#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <stdarg.h>

/* tokens */
#define T_NONE			-1
#define T_EOF			0
#define T_BRACKET_OPEN		1
#define T_BRACKET_CLOSE		2
#define T_PLUS_MINUS		3
#define T_STRING		10
#define T_NUMBER		11
#define T_IDENTIFIER		100
#define T_TO			101
#define T_DATETIME		200
#define T_INTERVAL		201
#define T_DATETIME_IV_FQUAL	202
#define T_OTHER			999

int savedToken = T_NONE;

/* these consitute one token of tpe T_OTHER */
char *op2Tab[] = {
	"||",
	"::",
	".*",
	"<=",
	">=",
	"<>",
	"!=",
	"==",
	NULL
};

struct IdTab {
	char *name;
	int token;
};

struct IdTab idTab[] = {
	"datetime",	T_DATETIME,
	"interval",	T_INTERVAL,
	"to",		T_TO,
	"year",		T_DATETIME_IV_FQUAL,
	"month",	T_DATETIME_IV_FQUAL,
	"day",		T_DATETIME_IV_FQUAL,
	"hour",		T_DATETIME_IV_FQUAL,
	"minute",	T_DATETIME_IV_FQUAL,
	"second",	T_DATETIME_IV_FQUAL,
	"fraction",	T_DATETIME_IV_FQUAL,
	NULL,		T_NONE
};

FILE *ifp, *ofp;

#define MAXLINE 80	/* min output line length, \n after next token */

#define SQLTXTMAX (64*1024)
char buf[SQLTXTMAX];

char tokenBuf[SQLTXTMAX];
char putbackBuf[SQLTXTMAX];

int NextToken(char *pText);
void PutBackToken(int token, char *pText);
int ScanNumber(char **pNext, char *pText);
void ParseDatetimeInterval();
void Error(char *str, ...);

int
main(int argc, char **argv)
{
	int l, token;
	char *p;
	int plusMinus = 0;
	int lineLen = 0;

	if (argc < 2 || !strcmp(argv[1], "-"))
		ifp = stdin;
	else {
		ifp = fopen(argv[1], "r");
		if (!ifp)
			Error("cannot open input file %s", argv[1]);
	}

	if (argc < 3 || !strcmp(argv[2], "-"))
		ofp = stdout;
	else {
		ofp = fopen(argv[2], "w");
		if (!ofp)
			Error("cannot open input file %s", argv[2]);
	}

	l = fread(buf, 1, SQLTXTMAX - 1, ifp);
	if (ferror(ifp))
		Error("cannot read input file %s", argv[1]);
	buf[l] = 0;

	/* convert to lowercase */
	for (p = buf; *p; ++p)
		if (isupper(*p))
			*p = tolower(*p);

	/* parse and output statement */
	while ((token = NextToken(tokenBuf)) != T_EOF) {
		/* Informix SQL: "100 + - - - - 1" is 101 !! 
		 * Print first '+' or '-', skip following ones.
		 */
		if (token == T_PLUS_MINUS) {
			if (plusMinus)
				continue;
			plusMinus = 1;
		}
		else
			plusMinus = 0;

		switch (token) {
		    case T_STRING:
			strcpy(tokenBuf, "<string>");
		    	break;
		    case T_NUMBER:
			strcpy(tokenBuf, "<number>");
		    	break;
		    case T_DATETIME:
			ParseDatetimeInterval();
			strcpy(tokenBuf, "<datetime>");
		    	break;
		    case T_INTERVAL:
			ParseDatetimeInterval();
			strcpy(tokenBuf, "<interval>");
		    	break;
		}

		l = strlen(tokenBuf);
		if (lineLen > 0 && lineLen + l + 1 >= MAXLINE) {
			fprintf(ofp, "\n");
			lineLen = 0;
		}

		if (lineLen) {
			fprintf(ofp, " ");
			lineLen += 1;
		}

		fprintf(ofp, "%s", tokenBuf);
		lineLen += l;
	}
	if (lineLen)
		fprintf(ofp, "\n");

	return 0;
}

void
PutBackToken(int token, char *pText)
{
	savedToken = token;
	strcpy(putbackBuf, pText);
}

int 
NextToken(char *pText)
{
	static char *next = buf;
	char *p, *p2;
	int c, n, l;

	if (savedToken != T_NONE) {
		int t = savedToken;

		savedToken = T_NONE;
		strcpy(pText, putbackBuf);
		return t;
	}

	/* Skip all leading sequences of whitspace, {} amd -- comments */
	do {
		while (isspace(*next))
			++next;

		p = next;

		if (next[0] == '-' && next[1] == '-') {
			next += 2;
			while (*next && *next++ != '\n')
				;
		}

		if (*next == '{') {
			++next;
			while (*next && *next++ != '}')
				;
		}
	} while (p != next);
	
	if (!*next)
		return T_EOF;

	p = next;
	if (*p == '(') {
		*pText++ = *p++;
		*pText = 0;
		next = p;
		return T_BRACKET_OPEN;
	}
	else if (*p == ')') {
		*pText++ = *p++;
		*pText = 0;
		next = p;
		return T_BRACKET_CLOSE;
	}
	else if (*p == '\'' || *p == '\"') {
		c = *p;
		*pText++ = *p++;
		while (*p) {
			*pText++ = *p++;
			if (p[-1] == c)
				if (*p == c)
					*pText++ = *p++;
				else
					break;
		}
		*pText = 0;
		next = p;
		return T_STRING;
	}
	else if (isalpha(*p) || *p == '_') {
		p2 = pText;
		*pText++ = *p++;
		while (isalnum(*p) || *p == '_' || *p == '$')
			*pText++ = *p++;
		*pText = 0;
		next = p;

		for (n = 0; idTab[n].name; ++n)
			if (!strcmp(idTab[n].name, p2))
				return idTab[n].token;
				
		return T_IDENTIFIER;
	}
	else if (*p == '+' || *p == '-') {
		*pText++ = *p++;
		*pText = 0;
		next = p;
		return T_PLUS_MINUS;
	}
	else if (isdigit(*p)) {
		return ScanNumber(&next, pText);
	}
	else {
		for (n = 0; op2Tab[n]; ++n) {
			l = strlen(op2Tab[n]);
			if (!strncmp(p, op2Tab[n], l))
				break;
		}
		if (op2Tab[n]) {
			strcpy(pText, op2Tab[n]);
			next = p + l;
			return T_OTHER;
		}

		*pText++ = *p++;
		*pText = 0;
		next = p;
		return T_OTHER;
	}

	Error("NextToken end");
	return T_EOF; /* not reached */
}

int
ScanNumber(char **pNext, char *pText)
{
	char *p = *pNext;

	while (isdigit(*p))
		*pText++ = *p++;

	if (*p == '.') {
		*pText++ = *p++;
		while (isdigit(*p))
			*pText++ = *p++;
	}

	if (*p == 'e' || *p == 'E') {
		*pText++ = *p++;
		if (*p == '-')
			*pText++ = *p++;
		else if (*p == '+')
			++p;
		while (isdigit(*p))
			*pText++ = *p++;
	}

	*pText = 0;
	*pNext = p;

	return T_NUMBER;
}

/* datetime and interval syntax:
 *
 * 	{T_DATETIME | T_INTERVAL} T_BRACKET_OPEN ... T_BRACKET_CLOSE \
 *   	    T_DATETIME_IV_FQUAL [prec] T_TO T_DATETIME_IV_FQUAL [prec]
 *
 *      prec: T_BRACKET_OPEN T_NUMBER T_BRACKET_CLOSE
 */
void
ParseDatetimeInterval()
{
	int token;

	if (NextToken(tokenBuf) != T_BRACKET_OPEN)
		return;

	while ((token = NextToken(tokenBuf)) != T_BRACKET_CLOSE)
		if (token == T_EOF)
			return;

	if (NextToken(tokenBuf) != T_DATETIME_IV_FQUAL)
		return;

	token = NextToken(tokenBuf);
	if (token == T_BRACKET_OPEN) {
		if (NextToken(tokenBuf) != T_NUMBER)
			return;
		if (NextToken(tokenBuf) != T_BRACKET_CLOSE)
			return;
		token = NextToken(tokenBuf);
	}

	if (token != T_TO)
		return;

	if (NextToken(tokenBuf) != T_DATETIME_IV_FQUAL)
		return;

	token = NextToken(tokenBuf);
	if (token == T_BRACKET_OPEN) {
		if (NextToken(tokenBuf) != T_NUMBER)
			return;
		if (NextToken(tokenBuf) != T_BRACKET_CLOSE)
			return;
	}
	else
		PutBackToken(token, tokenBuf);
}

void
Error(char *str, ...)
{
        va_list ap;

        va_start(ap, str);
        vfprintf(stderr, str, ap);
	va_end(ap);
        fprintf( stderr, "\n" );

        exit(1);
}
